BUILD_BIT='64bit'
BUILD_CPU='Intel(R) Core(TM) i7-6700K CPU @ 4.00GHz'
BUILD_DATE='2024-04-25'
BUILD_MACHINE='x86_64'
BUILD_OS='Linux Fedora 39'
BUILD_TIME='15:49'
BUILD_TYPE=''
BUILT_BY='antoine'
BUILT_ON='fedora'
COMPILER_VERSION='gcc (GCC) 13.2.1 20240316 (Red Hat 13.2.1-7)'
CYTHON_VERSION='3.0.6'
LINKER_VERSION='GNU ld version 2.40-14.fc39'
NVCC_VERSION=''
PYTHON_VERSION='3.12.2'
lib_gobjectintrospection10='1.78.1'
lib_gtk30='3.24.41'
lib_gtkx1130='3.24.41'
lib_nvenc='12'
lib_nvfbc='8.0'
lib_py3cairo='1.25.1'
lib_pygobject30='3.46.0'
lib_python3='3.12'
lib_python_Cython='3.0.10'
lib_python_PyGObject='3.44.1'
lib_python_PyOpenGL='3.1.7'
lib_python_PyOpenGLaccelerate='3.1.7'
lib_python_numpy='1.26.4'
lib_python_nvidiamlpy='12.535.133'
lib_python_pillow='10.3.0'
lib_python_pycairo='1.23.0'
lib_python_pytools='2023.1.1'
lib_python_pyxdg='0.28'
lib_python_setuptools='68.2.2'
lib_vpx='1.13.1'
lib_x11='1.8.9'
lib_x264='0.164.3108'
lib_xcomposite='0.4.5'
lib_xdamage='1.1.5'
lib_xext='1.3.5'
lib_xfixes='6.0.0'
lib_xkbfile='1.1.1'
lib_xrandr='1.5.4'
lib_xtst='1.2.4'
